/*
 * ClassFinder - Finding the names of all classes in the class path
 * Copyright (C) 2006  Johannes Mayer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package de.uulm.plugin;

import java.util.*;
import java.util.jar.*;
import java.io.File;
import java.io.IOException;


/**
* Searches for all classes in the class path. The names and also objects of
* these classes can be retrieved.
*
* @author  Johannes Mayer
* @version 0.1.2, 2006-12-19
*/
@SuppressWarnings("unchecked")
public class ClassFinder {

    /**
    * There should be no object of this tool class.
    */
    private ClassFinder() {
       
    }

    /**
    * Returns a list of <i>Class</i> objects of all classes in the class path.
    *
    * @return	a list of <i>Class</i> objects of all classes in the class path.
    */
    public static List<Class> getClasses() {
	List<Class> classes = new Vector<Class>();

	for (String className : getClassNames()) {
	    try {
        
		classes.add(Class.forName(className));
	    }
	    // not sufficient, since other exceptions (such as NoClassDefError)
	    // may occur!
            //catch (ClassNotFoundException ex) {
            //}
	    // :(
            catch (Throwable ex) {
            }
	}

	return classes;
    }

    /**
    * Returns a list of names of all classes in the class path.
    *
    * @return	a list of names of all classes in the class path.
    */
    public static List<String> getClassNames() {
	List<String> classNames = new Vector<String>();
	load(classNames);
	return classNames;
    }

    /**
    * Loads the names of all classes in the class path and stores them
    * in <i>classNames</i>.
    */
    private static void load(List<String> classNames) {
	// retrieve the path separator character
	String sep = System.getProperty("path.separator");
	if (sep.length() > 1)
	    new RuntimeException("path.separator is not a single character");

	// retrieve the class path
	String classPath = System.getProperty("java.class.path");
        
	// split the class path up into its elements
	Map<String,Object> m = new HashMap<String,Object>();
	StringTokenizer st = new StringTokenizer(classPath, sep);
	while (st.hasMoreTokens()) {
	    String elem = st.nextToken();
	    m.put(elem, "");
	}

	// add the root of current thread's class loader also to the list of roots;
	// circumvents a problem when being used in a web application
	// (solution contributed by Tobias Haf)
        try {
	    ClassLoader cl = Thread.currentThread().getContextClassLoader();
            String root = cl.getResource("/").getFile();
	    m.put(root, "");
        }
        catch(NullPointerException e) {    
        }

	// iterate over all path elements
	for (String elem : m.keySet()) {
	    if (elem.endsWith(".jar"))
		loadJar(classNames, elem);
	    else {
		File directory = new File(elem);
		if (directory.exists() && directory.isDirectory())
		    loadDirectory(classNames, "", directory);
	    }
	}
    }

    /**
    * Loads the names of all classes in the jar file <i>jarFile</i> and stores them
    * in <i>classNames</i>.
    */
    private static void loadJar(List<String> classNames, String jarFile) {
        try {
	    JarFile f = new JarFile(jarFile);
	    for (Enumeration e = f.entries(); e.hasMoreElements();) {
	        JarEntry entry = (JarEntry) e.nextElement();
	        String name = entry.getName();
	        if (name.endsWith(".class")) {
		    name = name.substring(0, name.length()-6);
		    name = name.replace("/", ".");
		    // add name as new class name
		    classNames.add(name);
	        }
	    }
	}
	catch (IOException ex) {
	}
    }

    /**
    * Loads the names of all classes in the directory <i>dir</i> and stores them
    * in <i>classNames</i>. The string <i>prefix</i> contains the package name being
    * the prefix of all class names. The search is also performed in sub-directories.
    */
    private static void loadDirectory(List<String> classNames, String prefix, File dir) {
        // get the files contained in the directory
        File[] files = dir.listFiles();
        for (int i = 0; i < files.length; i++) {
	    String name = files[i].getName();
	    // we are only interested in .class files
	    if (name.endsWith(".class")) {
	        // removes the .class extension
	        String className = prefix + name.substring(0, name.length()-6);
		// add className as new class name
		classNames.add(className);
	    }
	    // recursively search within all subpackages
	    // if recursive is true
	    else if (files[i].isDirectory())
	        loadDirectory(classNames, prefix + files[i].getName() + ".", files[i]);
	}
    }

}
