/*
 * PluginRegistry - Registry making all installed plugins accessible
 * Copyright (C) 2006  Johannes Mayer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package de.uulm.plugin;

import java.util.*;
import java.lang.reflect.Array;

/**
* Searches for plugin classes (i.e. classes implementing the
* <i>Plugin</i> marker interface) provides objects of these classes.
*
* @author  Johannes Mayer
* @version 0.1, 2006-12-05
*/
@SuppressWarnings("unchecked")
public class PluginRegistry {

    /**
    * Stores the <i>Class</i> objects of all plugins which have already
    * been loaded. In case of <i>Singleton</i> plugins, the only instance
    * of this plugin class is stored as soon as one has been requested.
    * For non-singleton plugins, no instance is stored.
    */
    private static Map<Class,Plugin> plugins = new HashMap<Class,Plugin>();

    /**
    * Returns an instance of the given plugin class.
    */
    private static Plugin instantiate(Class clazz) {
	try {
	    return (Plugin) clazz.newInstance();
	}
	catch (InstantiationException ex) {
	}
	catch (IllegalAccessException ex) {
	}
	return null;
    }

    /**
    * Again searches the whole class path for newly added plugins.
    */
    private static synchronized void reload() {
	List<Class> cs = ClassFinder.getClasses();
	for (Class c : cs)
	    if (!plugins.containsKey(c) && Plugin.class.isAssignableFrom(c) && !c.isInterface())
		plugins.put(c, null);
    }

    /**
    * The only instance of the plugin registry.
    */
    private static PluginRegistry reg = null;
    
    /**
    * Returns the only instance of the plugin registry.
    *
    * @return	the only instance of the plugin registry.
    */
    public static PluginRegistry getInstance() {
	if (reg == null)
	    reg = new PluginRegistry();
	return reg;
    }

    /**
    * Objects can only be constructed from within this class in order
    * to ensure that this class implements the Singleton pattern.
    */
    private PluginRegistry() {
	reload();
    }

    /**
    * Initiate the search for newly added plugins.
    */
    public void load() {
	reload();
    }
    
    /**
    * Returns all plugins that are a sub type of <i>pluginType</i>. For plugins
    * that implement the <i>Singleton</i> interface, there is at most one
    * instance which is returned each time. For other plugins, each time a new
    * instance is constructed and returned.
    *
    * @param	pluginType	the super type of all plugins to be retrieved
    * @return	all plugins that are a sub type of <i>pluginType</i>
    */
    public Plugin[] getPlugins(Class pluginType) {
    	List<Plugin> ps = new Vector<Plugin>();
	for (Class c : plugins.keySet())
	    if (pluginType.isAssignableFrom(c)) {
		Plugin p = null;
	    	if (Singleton.class.isAssignableFrom(c)) {
		    // retrieve instance
		    p = plugins.get(c);
		    if (p == null) {
		    	// no instance => construct instance
		        p = instantiate(c);
			plugins.put(c, p);
		    }
		}
		else {
		    // always construct new instance
		    p = instantiate(c);
		}
		if (p != null)
		    ps.add(p);
	    }

	Plugin[] array = (Plugin[]) Array.newInstance(pluginType, ps.size());
	return ps.toArray(array);
    }

}
